import os
import numpy as np
import pandas as pd
import mavenn
from tensorflow import keras
from tensorflow.keras.regularizers import l1_l2



class SurrogateBase():
    """
    Base class for surrogate model.
    """
    def __init__(self):
        pass

    def train(self, x):
        raise NotImplementedError()

    def get_params(self, x):
        raise NotImplementedError()


class SurrogateLinear(SurrogateBase):
    """Module for linear surrogate model (no GE or noise models).

    Parameters
    ----------
    l1 : float, optional
        L1 regularization penalty
    l2 : float, optional
        L2 regularization penalty

    Returns
    -------
    keras.Model  
    """
    def __init__(self, input_shape, num_tasks, l1=1e-8, l2=1e-4,
                 alphabet=['A','C','G','T']):

        self.model = self.build(input_shape, num_tasks, l1, l2)
        self.alphabet = alphabet

    def build(self, input_shape, num_tasks, l1, l2):
        """Build linear surrogate model."""
        
        N,L,A = input_shape

        # input layer
        inputs = keras.layers.Input(shape=(L,A))
        flatten = keras.layers.Flatten()(inputs)
        outputs = keras.layers.Dense(num_tasks,
                                     activation='linear',
                                     kernel_regularizer=l1_l2(l1=l1, l2=l2),
                                     use_bias=True)(flatten)

        # compile model
        return keras.Model(inputs=inputs, outputs=outputs)


    def train(self, x, y, learning_rate=1e-3, epochs=500, batch_size=100, early_stopping=True,
              patience=25, restore_best_weights=True, rnd_seed=None, save_dir=None, verbose=1):
        """Train linear surrogate model.
        
        """
        # generate data splits
        train_index, valid_index, test_index = data_splits(x.shape[0], test_split=0.1, valid_split=0.1, rnd_seed=rnd_seed)
        x_train = x[train_index]
        y_train = y[train_index]
        x_valid = x[valid_index]
        y_valid = y[valid_index]
        x_test = x[test_index]
        y_test = y[test_index]

        # set up optimizer and metrics
        self.model.compile(keras.optimizers.Adam(learning_rate), loss='mse')

        # early stopping callback
        es_callback = keras.callbacks.EarlyStopping(monitor='val_loss',
                                                    patience=patience,
                                                    verbose=1,
                                                    mode='min',
                                                    restore_best_weights=restore_best_weights)

        # reduce learning rate callback
        reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='val_loss',
                                                      factor=0.2,
                                                      patience=3,
                                                      min_lr=1e-7,
                                                      mode='min',
                                                      verbose=verbose)

        # fit model to data
        history = self.model.fit(x_train, y_train,
                                 epochs=epochs,
                                 batch_size=batch_size,
                                 shuffle=True,
                                 validation_data=(x_valid, y_valid),
                                 callbacks=[es_callback, reduce_lr],
                                 verbose=verbose)
        
        if save_dir is not None:
            self.model.save(os.path.join(save_dir, 'linear_model'))

        return (self.model, None)


    def get_params(self, gauge=None, save_dir=None):
        """Get parameters of the model.
        """
        for layer in self.model.layers:
            weights = layer.get_weights()

        return self.model.layers[2].get_weights()[0]
    

    def get_logo(self, full_length=None, mut_window=None):

        # insert the (potentially-delimited) additive logo back into the max-length sequence
        if full_length is None:
            full_length = self.L
        #additive_logo = self.theta_dict['logomaker_df']
        additive_logo = self.get_params(self.model)
        additive_logo.fillna(0, inplace=True) #if necessary, set NaN parameters to zero
        if mut_window is not None:
            additive_logo_zeros = np.zeros(shape=(full_length, self.A))
            additive_logo_zeros[mut_window[0]:mut_window[1], :] = additive_logo
            additive_logo = additive_logo_zeros

        return additive_logo



class SurrogateMAVENN(SurrogateBase):
    """Module for MAVE-NN surrogate models (optional GE and noise models).

    Parameters
    ----------
    gpmap : string {'additive' or 'pairwise'}
        Define MAVE-NN surrogate model used to interpret deep learning model.
            'additive'  :   Assume that each position contributes independently to the latent phenotype.
            'pairwise'  :   Assume that every pair of positions contribute to the latent phenotype.
    regression_type : string
        Type of regression used for measurement process.
            'MPA'   :   measurement process agnostic (categorical y-values).
            'GE'    :   global epistasis (continuous y-values).
    linearity : string
        Define use of additional nonlinearity for fitting data.
            'nonlinear' :   Additionally fit data using GE nonlinear function.
            'linear'    :   Do not apply GE nonlinearity for fitting data.
    noise : string
        Noise model to use for when defining a GE model (no effect on MPA models).
        See https://mavenn.readthedocs.io/en/latest/math.html for more info.
            'Gaussian'  :   Gaussian-based noise model.
            'Cauchy'    :   Cauchy-based noise model.
            'SkewedT'   :   SkewedT-based noise model.
    noise_order : int
        In the GE context, the order of the polynomial(s) used to define noise model parameters.
        In the linear context, the order is zero by default.
    reg_strength : float
        L2 regularization strength for G-P map parameters.
    alphabet : list
        The alphabet used to determine the C characters in the logo such that
        each entry is a string; e.g., ['A','C','G','T'] for DNA.
    deduplicate : boole
        Remove duplicate sequence-function pairs in dataset (default: True).
    gpu : boolean
        Enable GPUs (default: True).

    Returns
    -------
    keras.Model
    """
    def __init__(self, input_shape, num_tasks, gpmap='additive', regression_type='GE',
                 linearity='nonlinear', noise='SkewedT', noise_order=2, reg_strength=0.1,
                 alphabet=['A','C','G','T'], deduplicate=True, gpu=True):
        
        self.N, self.L, self.A = input_shape
        self.num_tasks = num_tasks
        self.gpmap = gpmap
        self.regression_type = regression_type
        self.linearity = linearity
        self.noise = noise
        if self.linearity == 'linear':
            self.noise_order = 0
        elif self.linearity == 'nonlinear':
            self.noise_order = noise_order
        self.reg_strength = reg_strength
        self.alphabet = alphabet
        self.deduplicate = deduplicate
        self.gpu = gpu
    

    def dataframe(self, x, y, alphabet, gpu):
        
        N = x.shape[0]
        mave_df = pd.DataFrame(columns = ['y', 'x'], index=range(N))
        mave_df['y'] = y
        
        if gpu is False:
            alphabet_dict = {}
            idx = 0
            for i in range(len(alphabet)):
                alphabet_dict[i] = alphabet[i]
                idx += 1
            for i in range(N): #standard approach
                seq_index = np.argmax(x[i,:,:], axis=1)
                seq = []
                for s in seq_index:
                    seq.append(alphabet_dict[s])
                seq = ''.join(seq)
                mave_df.at[i, 'x'] = seq

        elif gpu is True: #convert entire matrix at once (~twice as fast as standard approach if running on GPUs)
            seq_index_all = np.argmax(x, axis=-1)
            num2alpha = dict(zip(range(0, len(alphabet)), alphabet))
            seq_vector_all = np.vectorize(num2alpha.get)(seq_index_all)
            seq_list_all = seq_vector_all.tolist()
            dictionary_list = []
            for i in range(0, N, 1):
                dictionary_data = {0: ''.join(seq_list_all[i])}
                dictionary_list.append(dictionary_data)
            mave_df['x'] = pd.DataFrame.from_dict(dictionary_list)

        return mave_df

    
    def train(self, x, y, learning_rate=5e-4, epochs=500, batch_size=100,
              early_stopping=True, patience=25, restore_best_weights=True,
              save_dir=None, verbose=1):

        # convert matrix of one-hots into sequence dataframe
        if verbose:
            print('  Creating sequence dataframe...')
            print('')
        mave_df = self.dataframe(x, y, alphabet=self.alphabet, gpu=self.gpu)
        if verbose:
            print(mave_df)

        if self.deduplicate:
            mave_df.drop_duplicates(['y', 'x'], inplace=True, keep='first')

        # ensure proper format
        mave_df['y'] = mave_df['y'].apply(lambda x: np.asarray(x).astype('float32'))
 
        # split dataset for MAVE-NN
        mave_df['set'] = np.random.choice(a=['training','test','validation'],
                                        p=[.6,.2,.2],
                                        size=len(mave_df))
        new_cols = ['set'] + list(mave_df.columns[0:-2]) + ['x']
        mave_df = mave_df[new_cols]
        if save_dir is not None:
            mave_df.to_csv(os.path.join(save_dir, 'mave_df.csv'),  index=False) #.csv.gz', compression='gzip')

        trainval_df, self.test_df = mavenn.split_dataset(mave_df)
        self.y_cols = trainval_df.columns[1:-1] #get the column index for the counts
        
        # show dataset sizes
        if verbose:
            print(f'Train + val set size : {len(trainval_df):6,d} observations')
            print(f'Test set size        : {len(self.test_df):6,d} observations')
            print('')

        gpmap_kwargs = {'L': self.L,
                        'C': self.A,
                        'theta_regularization': self.reg_strength}
        
        if verbose:
            if self.gpmap == 'additive':
                print('Initializing additive model... %s parameters' % int(self.L*4))
            elif self.gpmap == 'neighbor':
                print('Initializing neighbor model... %s parameters' % int((self.L-1)*16))
            elif self.gpmap == 'pairwise':
                print('Initializing pairwise model... %s parameters' % int((self.L*(self.L-1)*16)/2))
            print('')
        
        # create the model
        if self.regression_type == 'MPA':
            self.model = mavenn.Model(L=self.L,
                                Y=len(self.y_cols),
                                alphabet=self.alphabet,
                                ge_nonlinearity_type=self.linearity,
                                regression_type='MPA',
                                gpmap_type=self.gpmap, 
                                gpmap_kwargs=gpmap_kwargs)
            y_data = trainval_df[self.y_cols]
            
        elif self.regression_type == 'GE':
            self.model = mavenn.Model(L=self.L,
                                Y=len(self.y_cols),
                                alphabet=self.alphabet,
                                ge_nonlinearity_type=self.linearity,
                                regression_type='GE',
                                ge_noise_model_type=self.noise,
                                ge_heteroskedasticity_order=self.noise_order, 
                                gpmap_type=self.gpmap,
                                gpmap_kwargs=gpmap_kwargs)
            y_data = trainval_df['y']
        
        # set training data
        self.model.set_data(x=trainval_df['x'], y=y_data,
                        validation_flags=trainval_df['validation'],
                        shuffle=True)
        
        # fit model to data
        self.model.fit(learning_rate=learning_rate,
                epochs=epochs,
                batch_size=batch_size,
                early_stopping=early_stopping,
                early_stopping_patience=patience,
                linear_initialization=False,
                #restore_best_weights=self.restore_best_weights,
                verbose=False)
        
        if save_dir is not None:
            self.model.save(os.path.join(save_dir, 'mavenn_model_%s' % self.gpmap))

        return (self.model, mave_df)


    def get_info(self, save_dir=None, verbose=1):
        """Function to return estimated variational information from MAVE-NN model.

        Parameters
        ----------
        save_dir : str
            Directory for saving figures to file.
        verbose : bool
            print info
        Returns
        -------
        I_pred : float
            MAVE-NN estimated variational information (I_pred), in bits.
        """
        # compute predictive information on test data
        if self.regression_type == 'MPA':
            I_pred, dI_pred = self.model.I_predictive(x=self.test_df['x'], y=self.test_df[self.y_cols])
        elif self.regression_type == 'GE':
            I_pred, dI_pred = self.model.I_predictive(x=self.test_df['x'], y=self.test_df['y'])

        if verbose:
            print('')
            print('Model performance:')
            print(f'  test_I_pred: {I_pred:.3f} +- {dI_pred:.3f} bits')
            print('  max I_var:', np.amax(self.model.history['I_var']))
            print('  max val_I_var:', np.amax(self.model.history['val_I_var']))
            print('')

        if save_dir is not None:
            # save information content to text file
            if os.path.exists(os.path.join(save_dir, 'model_info.txt')):
                os.remove(os.path.join(save_dir,'model_info.txt'))
            f_out = open(os.path.join(save_dir, 'model_info.txt'),'w')
            print(f'test_I_pred: {I_pred:.3f} bits', file=f_out)
            print('max I_var:',np.amax(self.model.history['I_var']), file=f_out)
            print('max val_I_var:',np.amax(self.model.history['val_I_var']), file=f_out)
            print('', file=f_out)
            f_out.close()
        
        return I_pred


    def get_params(self, gauge='empirical', save_dir=None):
        """Function to return trained parameters from MAVE-NN model.

        Parameters
        ----------
        gauge : gauge mode used to fix model parameters.
                See https://mavenn.readthedocs.io/en/latest/math.html for more info.
            'uniform'   :   hierarchical gauge using a uniform sequence distribution over
                            the characters at each position observed in the training set
                            (unobserved characters are assigned probability 0).
            'empirical' :   uses an empirical distribution computed from the training data.
            'consensus' :   wild-type gauge using the training data consensus sequence.
        save_dir : str
            Directory for saving figures to file.
        
        Returns
        -------
        tuple 
            theta_0     :   float
                Constant term in trained parameters.
            theta_lc    :   numpy.ndarray
                Additive terms in trained parameters (shape : (L,C)).
            theta_lclc  :   numpy.ndarray
                Pairwise terms in trained parameters (shape : (L,C,L,C)), if gpmap is 'pairwise'.
        """
        # fix gauge mode for model representation
        self.theta_dict = self.model.get_theta(gauge=gauge) #for usage: theta_dict.keys()
        
        theta_0 = self.theta_dict['theta_0']
        theta_lc = self.theta_dict['theta_lc']
        theta_lc[np.isnan(theta_lc)] = 0
        if save_dir is not None:
            np.save(os.path.join(save_dir, 'theta_0.npy'), theta_0)
            np.save(os.path.join(save_dir, 'theta_lc.npy'), theta_lc)

        if self.gpmap == 'pairwise':
            theta_lclc = self.theta_dict['theta_lclc']
            theta_lclc[np.isnan(theta_lclc)] = 0
            if save_dir is not None:
                np.save(os.path.join(save_dir, 'theta_lclc.npy'), theta_lclc)
        else:
            theta_lclc = None

        return (theta_0, theta_lc, theta_lclc)


    def get_logo(self, full_length=None, mut_window=None):
        """Function to place trained additive parameters into surrounding 
        nonmutated sequence (zeros).

        Parameters
        ----------
        full_length : int
            Full length of sequence.
        mut_window : [int, int]
            Index of start and stop position along sequence to probe;
            i.e., [start, stop], where start < stop and both entries
            satisfy 0 <= int <= L.

        Returns
        -------
        additive_logo : numpy.ndarray
            Additive logo parameters (shape : ('full_length',C)).
        """
        # insert the (potentially-delimited) additive logo back into the max-length sequence
        if full_length is None:
            full_length = self.L
        additive_logo = self.theta_dict['logomaker_df']
        additive_logo.fillna(0, inplace=True) #if necessary, set NaN parameters to zero
        if mut_window is not None:
            additive_logo_zeros = np.zeros(shape=(full_length, self.A))
            additive_logo_zeros[mut_window[0]:mut_window[1], :] = additive_logo
            additive_logo = additive_logo_zeros

        return additive_logo


def data_splits(N, test_split, valid_split, rnd_seed=None):
    """Function to determine which sequences randomly split into train, validation, test set.

    Parameters
    ----------
    N : int
        number of data
    test_splitc : float (between 0 and 1)
        percent to split into test set
    valid_split : float (between 0 and 1)
        percent to split into validation set
    rnd_seed : int
        random number seed

    Returns
    -------
    train_index
        array of indices to be included in training set
    valid_index
        array of indices to be included in validation set
    test_index
        array of indices to be included in test set
    """
    if rnd_seed:
        np.random.seed(rnd_seed)
    train_split = 1 - test_split - valid_split
    shuffle = np.random.permutation(range(N))
    num_valid = int(valid_split*N)
    num_test = int(test_split*N)
    test_index = shuffle[:num_test]
    valid_index = shuffle[num_test:num_test+num_valid]
    train_index = shuffle[num_test+num_valid:]
    return train_index, valid_index, test_index





import numpy as np
from tqdm import tqdm


class BasePredictor():
    """
    Base class for running inference on in silico mutated sequences.
    """

    def __init__(self):
        raise NotImplementedError()

    def __call__(self, x):
        raise NotImplementedError()



class ScalarPredictor(BasePredictor):
    """Module for handling scalar-based model predictions.

    Parameters
    ----------
    pred_fun : built-in function
        Function for returning model predictions.
    task_idx : int
        Task index corresponding to a specific output head.
    batch_size : int
        The number of predictions per batch.

    Returns
    -------
    torch.Tensor
        Batch of scalar predictions corresponding to inputs.
    """

    def __init__(self, pred_fun, task_idx=0, batch_size=64, **kwargs):
        self.pred_fun = pred_fun
        self.task_idx = task_idx
        self.kwargs = kwargs
        self.batch_size = batch_size

    def __call__(self, x):
        pred = predict_in_batches(x, self.pred_fun, self.batch_size, **self.kwargs)
        return pred[self.task_idx]



class ProfilePredictor(BasePredictor):
    """Module for handling profile-based model predictions.

    Parameters
    ----------
    pred_fun : function
        Function for returning model predictions.
    task_idx : int
        Task index corresponding to a specific output head.
    batch_size : int
        The number of predictions per batch.
    reduce_fun : function
        Function for reducing profile prediction to scalar.

    Returns
    -------
    torch.Tensor
        Batch of scalar predictions corresponding to inputs.
    """

    def __init__(self, pred_fun, task_idx=0, batch_size=64, reduce_fun=np.sum, save_dir=None, **kwargs):
        self.pred_fun = pred_fun
        self.task_idx = task_idx
        self.batch_size = batch_size
        self.reduce_fun = reduce_fun
        #self.axis = axis
        self.save_dir = save_dir
        self.kwargs = kwargs

    def __call__(self, x):
        # get model predictions (all tasks)
        pred = predict_in_batches(x, self.pred_fun, self.batch_size, **self.kwargs)

        # reduce profile to scalar across axis for a given task_idx
        pred = self.reduce_fun(pred[:,:,self.task_idx], save_dir=self.save_dir)
        return pred[:,np.newaxis]



'''
class BPNetPredictor(BasePredictor):

    def __init__(self, pred_fun, task_idx=0, batch_size=64, reduce_fun=np.sum, axis=1, strand='pos', **kwargs):
        self.pred_fun = pred_fun
        self.task_idx = task_idx
        self.batch_size = batch_size
        self.reduce_fun = reduce_fun
        self.axis = axis
        self.kwargs = kwargs
        if strand == 'pos':
            self.strand = 0
        else:
            self.strand = 1

    def __call__(self, x):

        # get model predictions (all tasks)
        pred = predict_in_batches(x, self.pred_fun, self.batch_size, **self.kwargs)

        # reduce bpnet profile prediction to scalar across axis for a given task_idx
        pred = pred[self.task_idx][0][:,self.strand]
        pred = self.reduce_fun(pred, axis=self.axis)
        return pred[:,np.newaxis]
'''



'''
class CustomPredictor():

    def __init__(self, pred_fun, reduce_fun, task_idx, batch_size):
        self.pred_fun = pred_fun
        self.reduce_fun = reduce_fun
        self.task_idx = task_idx
        self.batch_size = batch_size

    def __call__(self, x):
        # code goes here: process predictions into scalar
        return predictions

'''

################################################################################
# useful functions
################################################################################


def predict_in_batches(x, model_pred_fun, batch_size=None, **kwargs):
    """Function to compute model predictions in batch mode.

    Parameters
    ----------
    x : torch.Tensor
        One-hot sequences (shape: (N, L, A)).
    model_pred_fun : function
        Built-in function for accessing model inference on inputs.
    batch_size : int
        The number of predictions per batch of model inference.

    Returns
    -------
    numpy.ndarray
        Model predictions.
    """

    N, L, A = x.shape
    num_batches = np.floor(N/batch_size).astype(int)
    pred = []
    for i in tqdm(range(num_batches), desc="Inference"):
        pred.append(model_pred_fun(x[i*batch_size:(i+1)*batch_size], **kwargs))
    if num_batches*batch_size < N:
        pred.append(model_pred_fun(x[num_batches*batch_size:], **kwargs))

    try:
        preds = np.concatenate(pred, axis=1)
    except ValueError as ve:
        preds = np.vstack(pred)
    return preds



def profile_sum(pred, save_dir=None):
    """Function to transform predictions to scalars using summation.

    Parameters
    ----------
    pred : np.ndarray
        Batch of profile-based predictions.

    Returns
    -------
    numpy.ndarray
        Batch of scalar predictions.
    """

    sum = np.sum(pred, axis=1)
    return sum


def profile_pca(pred, save_dir=None):
    """Function to transform predictions to scalars using principal component analysis (PCA).

    Parameters
    ----------
    pred : np.ndarray
        Batch of profile-based predictions.

    Returns
    -------
    numpy.ndarray
        Batch of scalar predictions formed from projection of embedded
        profiles onto the first principal component.
    """

    N, B = pred.shape #B : number of bins in profile
    Y = pred.copy()
    sum = np.sum(pred, axis=1) #needed for sense correction

    # normalization: mean of all distributions is subtracted from each distribution
    mean_all = np.mean(Y, axis=0)
    for i in range(N):
        Y[i,:] -= mean_all

    u,s,v = np.linalg.svd(Y.T, full_matrices=False)
    vals = s**2 #eigenvalues
    vecs = u #eigenvectors
    
    U = Y.dot(vecs)
    v1, v2 = 0, 1
    
    corr = np.corrcoef(sum, U[:,v1])
    if corr[0,1] < 0: #correct for eigenvector "sense"
        U[:,v1] = -1.*U[:,v1]

    #impress.plot_eig_vals(vals, save_dir=save_dir)
    #impress.plot_eig_vecs(U, v1=v1, v2=v2, save_dir=save_dir)

    return U[:,v1]


"""
def custom_reduce(pred):
    # code to reduce predictions to (N,1)
    return pred_reduce
"""





import numpy as np
from tqdm import tqdm


class BaseMutagenesis:
    """
    Base class for in silico MAVE data generation for a given sequence.
    """
    def __call__(self, x, num_sim):
        """Return an in silico MAVE based on mutagenesis of 'x'.

        Parameters
        ----------
        x : torch.Tensor
            one-hot sequence (shape: (L, A)).
        num_sim : int
            Number of sequences to mutagenize.

        Returns
        -------
        torch.Tensor
            Batch of one-hot sequences with random augmentation applied.
        """
        raise NotImplementedError()


class RandomMutagenesis(BaseMutagenesis):
    """Module for performing random mutagenesis.

    Parameters
    ----------
    mut_rate : float, optional
        Mutation rate for random mutagenesis (defaults to 0.1).
    uniform : bool
        uniform (True), Poisson (False); sets the number of mutations per sequence.

    Returns
    -------
    torch.Tensor
        Batch of one-hot sequences with random mutagenesis applied.
    """

    def __init__(self, mut_rate, uniform=False):
        self.mut_rate = mut_rate
        self.uniform = uniform

    def __call__(self, x, num_sim):

        L, A = x.shape
        avg_num_mut = int(np.ceil(self.mut_rate*L))

        # get indices of nucleotides
        x_index = np.argmax(x, axis=1)

        # sample number of mutations for each sequence
        if self.uniform:
            num_muts = int(avg_num_mut*np.ones((num_sim,), dtype=int))
        else:
            num_muts = np.random.poisson(avg_num_mut, (num_sim, 1))[:,0]
        num_muts = np.clip(num_muts, 0, L)
        one_hot = apply_mut_by_seq_index(x_index, (num_sim,L,A), num_muts)
        return one_hot


"""
class CustomMutagenesis(BaseMutagenesis):
    def __init__(self, param1, param2):
        self.param1 = param1
        self.param2 = param2

    def __call__(self, x, num_sim):

        # code goes here
        return one_hot
"""


################################################################################
# useful functions
################################################################################


def apply_mut_by_seq_index(x_index, shape, num_muts):
    """Function to perform random mutagenesis.

    Parameters
    ----------
    x_index : np.ndarray
        Indices of wildtype sequence.
    shape : list
        Shape of MAVE array; i.e., (num_sim,L,A).
    num_muts : int
        Number of mutations per sequence.

    Returns
    -------
    torch.Tensor
        Batch of one-hot sequences with random mutagenesis applied.
    """

    num_sim, L, A = shape
    one_hot = np.zeros((num_sim, L, A))

    # loop through and generate random mutagenesis
    for i, num_mut in enumerate(tqdm(num_muts, desc="Mutagenesis")):

        if i == 0: # keep wild-type sequence
            one_hot[i,:,:] = np.eye(A)[x_index]
        else:
            # generate mutation index
            mut_index = np.random.choice(range(0, L), num_mut, replace=False)

            # sample alphabet
            mut = np.random.choice(range(1, A), (len(mut_index)))

            # loop through sequence and add mutation index (note: up to 3 is added which does not map to [0,3] alphabet)
            seq_index = np.copy(x_index)
            for j, m in zip(mut_index, mut):
                seq_index[j] += m

            # wrap non-sensical indices back to alphabet -- effectively makes it random mutation
            seq_index = np.mod(seq_index, A)

            # create one-hot from index
            one_hot[i,:,:] = np.eye(A)[seq_index]
    return one_hot


import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import mavenn
import logomaker
from . import utils


def plot_y_hist(y_mut, save_dir=None):
    """Function for visualizing histogram of inferred predictions for MAVE dataset.

    Parameters
    ----------
    y_mut : numpy.ndarray
        Inferred predictions for sequences (shape: (N,1)).
    save_dir : str
        Directory for saving figures to file.

    Returns
    -------
    matplotlib.pyplot.Figure
    """

    # plot histogram of transformed deepnet predictions
    fig, ax = plt.subplots()
    ax.hist(y_mut, bins=100)
    ax.set_xlabel('y')
    ax.set_ylabel('Frequency')
    ax.axvline(y_mut[0], c='red', label='WT', linewidth=2, zorder=10) #wild-type prediction
    plt.legend(loc='upper right')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'mave_distribution.png'), facecolor='w', dpi=200)
        plt.close()
    #else:
        #plt.show()
    return fig 


def plot_performance(model, info, save_dir=None):
    """Function for visualizing MAVE-NN model training performance.

    Parameters
    ----------
    model : mavenn.src.model.Model
        MAVE-NN model object.
    info : float
        MAVE-NN estimated variational information (I_pred), in bits.
    save_dir : str
        Directory for saving figures to file.

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    
    # plot mavenn model performance
    fig, ax = plt.subplots(1, 1, figsize=[5, 5])
    # plot I_var_train, the variational information on training data as a function of epoch
    ax.plot(model.history['I_var'], label=r'I_var_train')
    # plot I_var_val, the variational information on validation data as a function of epoch
    ax.plot(model.history['val_I_var'], label=r'val_I_var')
    # plot I_pred_test, the predictive information of the final model on test data
    ax.axhline(info, color='C3', linestyle=':', label=r'test_I_pred')
    ax.set_xlabel('epochs')
    ax.set_ylabel('bits')
    ax.set_title('Training history')
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    ax.legend(loc='best')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'mavenn_training.png'), facecolor='w', dpi=200)
        plt.close()
    #else:
        #plt.show()
    return fig 


def plot_additive_logo(logo, center=True, view_window=None, alphabet=['A','C','G','T'], fig_size=None, save_dir=None):
    """Function for visualizing MAVE-NN additive model parameters.

    Parameters
    ----------
    logo : numpy.ndarray
        Additive model parameters (shape: (L,C)).
    center : bool
        If True, the stack of characters at each position will be centered
        around zero. This is accomplished by subtracting the mean value in
        each row of the matrix from each element in that row.
    view_window : [int, int]
        Index of start and stop position along sequence to probe;
        i.e., [start, stop], where start < stop and both entries
        satisfy 0 <= int <= L.
    alphabet : list
        The alphabet used to determine the C characters in the logo such that
        each entry is a string; e.g., ['A','C','G','T'] for DNA.
    fig_size : [float, float]
        Width, height in inches for matplotlib.pyplot figure.
    save_dir : str
        Directory for saving figures to file.

    Returns
    -------
    matplotlib.pyplot.Figure
    """

    # plot additive logo
    if fig_size is not None:
        fig, ax = plt.subplots(figsize=fig_size)
    else:
        fig, ax = plt.subplots(figsize=[10,3])

    if view_window is None:
        logo_fig = logo
    else:
        logo_fig = logo[view_window[0]:view_window[1]]
        logo_fig = utils.arr2pd(logo_fig, alphabet)

    logomaker.Logo(df=logo_fig,
                    ax=ax,
                    fade_below=.5,
                    shade_below=.5,
                    width=.9,
                    center_values=center,
                    #font_name='Arial Rounded MT Bold'  # causes excess warnings on colab
                    )
    if view_window is not None:
        ax.set_xticks(np.arange(0, view_window[1]-view_window[0], 1))
        ax.set_xticklabels(np.arange(view_window[0], view_window[1], 1))

    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.set_ylabel('Additive effect')
    ax.set_xlabel('Nucleotide position')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'additive_logo.png'), facecolor='w', dpi=200)
        plt.close()
    #else:
        #plt.show()
    return fig 


def plot_pairwise_matrix(theta_lclc, view_window=None, alphabet=['A','C','G','T'], save_dir=None):
    """Function for visualizing MAVE-NN pairwise model parameters.

    Parameters
    ----------
    theta_lclc : numpy.ndarray
        Pairwise model parameters (shape: (L,C,L,C)).
    view_window : [int, int]
        Index of start and stop position along sequence to probe;
        i.e., [start, stop], where start < stop and both entries
        satisfy 0 <= int <= L.
    alphabet : list
        The alphabet used to determine the C characters in the logo such that
        each entry is a string; e.g., ['A','C','G','T'] for DNA.
    save_dir : str
        Directory for saving figures to file.

    Returns
    -------
    matplotlib.pyplot.Figure
    """

    # plot maveen pairwise matrix
    fig, ax = plt.subplots(figsize=[10,5])
    ax, cb = mavenn.heatmap_pairwise(values=theta_lclc,
                                    alphabet=alphabet,
                                    ax=ax,
                                    gpmap_type='pairwise',
                                    cmap_size='2%',
                                    show_alphabet=False,
                                    cmap='seismic',
                                    cmap_pad=.1,
                                    show_seplines=True,            
                                    sepline_kwargs = {'color': 'k',
                                                      'linestyle': '-',
                                                      'linewidth': .5})
    if view_window is not None:
        ax.xaxis.set_ticks(np.arange(0, view_window[1]-view_window[0], 2))
        ax.set_xticklabels(np.arange(view_window[0], view_window[1], 2))  
    cb.set_label(r'Pairwise Effect',
                  labelpad=8, ha='center', va='center', rotation=-90)
    cb.outline.set_visible(False)
    cb.ax.tick_params(direction='in', size=20, color='white')
    ax.set_xlabel('Nucleotide position')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'pairwise_matrix.png'), facecolor='w', dpi=200)
        plt.close()
    #else:
        #plt.show()
    return fig 


def plot_y_vs_yhat(model, mave_df, save_dir=None):
    """Function for visualizing comparison of MAVE values and MAVE-NN predictions.

    Parameters
    ----------
    model : mavenn.src.model.Model
        MAVE-NN model object.
    mave_df : pandas.core.frame.DataFrame
        Dataframe containing MAVE training splits, y floats, and x strings  (shape : (N,3))
    save_dir : str
        Directory for saving figures to file.

    Returns
    -------
    matplotlib.pyplot.Figure
    """

    # plot mavenn y versus yhat
    fig, ax = plt.subplots(1,1,figsize=[5,5])
    trainval_df, test_df = mavenn.split_dataset(mave_df)
    y_test = test_df['y'] #get test data y values
    yhat_test = model.x_to_yhat(test_df['x']) #compute yhat on test data
    Rsq = np.corrcoef(yhat_test.ravel(), test_df['y'])[0, 1]**2 #compute R^2 between yhat_test and y_test    
    ax.scatter(yhat_test, y_test, color='C0', s=1, alpha=.1,
               label='test data')
    xlim = [min(yhat_test), max(yhat_test)]
    ax.plot(xlim, xlim, '--', color='k', label='diagonal', zorder=100)
    ax.set_xlabel('model prediction ($\hat{y}$)')
    ax.set_ylabel('measurement ($y$)')
    ax.set_title(f'Standard metric of model performance:\n$R^2$={Rsq:.3}');
    ax.legend(loc='upper left')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir,'mavenn_measure_yhat.png'), facecolor='w', dpi=200)
        plt.close()
    #else:
        #plt.show()
    return fig 
    
    
def plot_y_vs_phi(model, mave_df, save_dir=None):
    """Function for visualizing comparison of MAVE values and MAVE-NN latent phenotypes.

    Parameters
    ----------
    model : mavenn.src.model.Model
        MAVE-NN model object.
    mave_df : pandas.core.frame.DataFrame
        Dataframe containing MAVE training splits, y floats, and x strings  (shape : (N,3))
    save_dir : str
        Directory for saving figures to file.

    Returns
    -------
    matplotlib.pyplot.Figure
    """

    # plot mavenn y versus phi
    fig, ax = plt.subplots(1,1,figsize=[5,5])
    trainval_df, test_df = mavenn.split_dataset(mave_df)
    phi_test = model.x_to_phi(test_df['x']) #compute φ on test data
    phi_lim = [min(phi_test)-.5, max(phi_test)+.5] #set phi limit and create a grid in phi space
    phi_grid = np.linspace(phi_lim[0], phi_lim[1], 1000)
    yhat_grid = model.phi_to_yhat(phi_grid) #compute yhat each phi grid point
    q = [0.025, 0.975] #compute 95% CI for each yhat
    yqs_grid = model.yhat_to_yq(yhat_grid, q=q)
    ax.fill_between(phi_grid, yqs_grid[:, 0], yqs_grid[:, 1],
                   alpha=0.1, color='C1', lw=0, label='95% CI') #plot 95% confidence interval
    ax.plot(phi_grid, yhat_grid,
            linewidth=2, color='C1', label='nonlinearity') #plot GE nonlinearity
    y_test = test_df['y']
    ax.scatter(phi_test, y_test,
               color='C0', s=1, alpha=.1, label='test data',
               zorder=-100, rasterized=True) #plot scatter of φ and y values
    ax.set_xlim(phi_lim)
    ax.set_xlabel('latent phenotype ($\phi$)')
    ax.set_ylabel('measurement ($y$)')
    ax.set_title('GE measurement process')
    ax.legend(loc='upper left')
    fig.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'mavenn_measure_phi.png'), facecolor='w', dpi=200)
        plt.close()
    #else:
        #plt.show()
    return fig 


def plot_eig_vals(vals, save_dir=None):
    """Function for visualizing eigenvalue spectrum, if using PCA.

    Parameters
    ----------
    vals : numpy.ndarray
        Eigenvalues obtained from PCA.
    save_dir : str
        Directory for saving figures to file.

    Returns
    -------
    matplotlib.pyplot.Figure
    """

    fig, ax = plt.subplots()
    x = range(1,len(vals)+1)
    plt.scatter(x, vals)
    plt.title('Eigenvalue spectrum')
    plt.xlabel(r'$PC$')
    plt.ylabel(r'$\mathrm{\lambda}$', rotation=0)
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    plt.xlim([0,15])
    plt.ylim(-vals[0]/8., vals[0]+vals[0]/8.)
    plt.locator_params(nbins=15)
    plt.axhline(y=0, color='k', alpha=.5, linestyle='--', linewidth=1)
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'pca_eigvals.png'), facecolor='w', dpi=200)
        plt.close()
    #else:
        #plt.show()
    return fig 


def plot_eig_vecs(U, v1, v2, save_dir=None):
    """Function for visualizing subspace defined by two eigenvectors, if using PCA.

    Parameters
    ----------
    U : numpy.ndarray
        Eigenvectors obtained from PCA.
    v1 : int
        Index of first eigenvector.
    v2: int
        Index of second eigenvector.
    save_dir : str
        Directory for saving figures to file.

    Returns
    -------
    matplotlib.pyplot.Figure
    """

    fig, ax = plt.subplots()
    ax.scatter(U[:,v1], U[:,v2], s=1, facecolor='k', alpha=.5)
    ax.scatter(U[:,v1][0], U[:,v2][0], s=5, facecolor='r', label='WT') #first index is wild-type
    ax.set_xlabel(r'$PC_%s$' % (v1+1), fontsize=20)
    ax.set_ylabel(r'$PC_%s$' % (v2+1), fontsize=20)
    plt.legend(loc='best')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'pca_eigvecs.png'), facecolor='w', dpi=200)
        plt.close()
    #else:
        #plt.show()
    return fig


import numpy as np


class InSilicoMAVE():
    """Module for performing in silico MAVE.

    Parameters
    ----------
    mut_generator : class
        Module for performing random mutagenesis.
    pred_generator : class
        Module for inferring model predictions.
    seq_length : int
        Full length L of input sequence.
    mut_window : [int, int]
        Index of start and stop position along sequence to probe;
        i.e., [start, stop], where start < stop and both entries
        satisfy 0 <= int <= L.
    """
    def __init__(self, mut_generator, mut_predictor, seq_length, mut_window=None, context_agnostic=False):
        self.mut_generator = mut_generator
        self.mut_predictor = mut_predictor
        self.seq_length = seq_length
        self.mut_window = mut_window
        if mut_window is not None:
            self.start_position = mut_window[0]
            self.stop_position = mut_window[1]
        else:
            self.start_position = 0
            self.stop_position = seq_length
        self.context_agnostic = context_agnostic


    def generate(self, x, num_sim, seed=None, verbose=1):
        """Randomly mutate segments in a set of one-hot DNA sequences.

        Parameters
        ----------
        x : torch.Tensor
            One-hot sequence (shape: (L,A)).
        num_sim : int
            Number of sequences to mutagenize for in silico MAVE.
        seed : int, optional
            Sets the random number seed.

        Returns
        -------
        x_mut : numpy.ndarray
            Sequences simulated by mut_predictor.
        y_mut : numpy.ndarray
            Inferred predictions for sequences (shape: (N,1)).
        """
        if seed:
            np.random.seed(seed)
        if verbose:
            print('')
            print('Building in silico MAVE...')

        # generate in silico MAVE based on mutagenesis strategy
        if self.mut_window is not None:
            x_window = self.delimit_range(x, self.start_position, self.stop_position)
            x_mut = self.mut_generator(x_window, num_sim)
            if self.context_agnostic:
                x_mut_full = self.pad_seq_random(x_mut, x, self.start_position, self.stop_position)
            else:
                x_mut_full = self.pad_seq(x_mut, x, self.start_position, self.stop_position)
            y_mut = self.mut_predictor(x_mut_full)
        else:
            x_mut = self.mut_generator(x, num_sim)
            y_mut = self.mut_predictor(x_mut)

        return x_mut, y_mut


    def pad_seq(self, x_mut, x, start_position, stop_position):
        """Function to pad mutated sequences on both sides with the original unmutated context.
        
        Parameters
        ----------
        x_mut : numpy.ndarray
            Sequences with randomly mutated segments with length l < L
            defined by l = stop_position - start_position (shape: (N,l,C)).
        x : torch.Tensor
            Batch of one-hot sequences (shape: (L,A)).
        start_position : int
            Index of start position along sequence to probe.
        stop_position : int
            Index of stop position along sequence to probe.

        Returns
        -------
        numpy.ndarray
            Sequences with randomly mutated segments, padded to correct shape
            with random DNA (shape: (N,L,C)).
        """
        N = x_mut.shape[0]
        x = x[np.newaxis,:]
        x_start = np.tile(x[:,:start_position,:], (N,1,1))
        x_stop = np.tile(x[:,stop_position:,:], (N,1,1))
        return np.concatenate([x_start, x_mut, x_stop], axis=1)



    def pad_seq_random(self, x_mut, x, start_position, stop_position):
        """Function to pad mutated sequences on both sides with random DNA.
        
        Parameters
        ----------
        x_mut : numpy.ndarray
            Sequences with randomly mutated segments with length l < L
            defined by l = stop_position - start_position (shape: (N,l,C)).
        x : torch.Tensor
            Batch of one-hot sequences (shape: (L,A)).
        start_position : int
            Index of start position along sequence to probe.
        stop_position : int
            Index of stop position along sequence to probe.

        Returns
        -------
        numpy.ndarray
            Sequences with randomly mutated segments, padded to correct shape
            with random DNA (shape: (N,L,C)).
        """
        N = x_mut.shape[0]
        x_shuffle = dinuc_shuffle(x,num_shufs= N)
        return np.concatenate([x_shuffle[:,:start_position,:], x_mut, x_shuffle[:,stop_position:,:]], axis=1)



    def delimit_range(self, x, start_position, stop_position):
        """Function to delimit sequence to a specific region.
        
        Parameters
        ----------
        x : torch.Tensor
            Batch of one-hot sequences (shape: (L,A)).
        start_position : int
            Index of start position along sequence to probe.
        stop_position : int
            Index of stop position along sequence to probe.

        Returns
        -------
        numpy.ndarray
            Delimited sequences with length l < L defined by
            l = stop_position - start_position (shape: (N,l,C)).
        """
        return x[start_position:stop_position,:]
    



#------------------------------------------------------------------------------------
# Useful shuffle function
#------------------------------------------------------------------------------------


# taken from https://github.com/kundajelab/deeplift/blob/master/deeplift/dinuc_shuffle.py
def dinuc_shuffle(seq, num_shufs=None, rng=None):
    """
    Creates shuffles of the given sequence, in which dinucleotide frequencies
    are preserved.
    Arguments:
        `seq`: either a string of length L, or an L x D NumPy array of one-hot
            encodings
        `num_shufs`: the number of shuffles to create, N; if unspecified, only
            one shuffle will be created
        `rng`: a NumPy RandomState object, to use for performing shuffles
    If `seq` is a string, returns a list of N strings of length L, each one
    being a shuffled version of `seq`. If `seq` is a 2D NumPy array, then the
    result is an N x L x D NumPy array of shuffled versions of `seq`, also
    one-hot encoded. If `num_shufs` is not specified, then the first dimension
    of N will not be present (i.e. a single string will be returned, or an L x D
    array).
    """

    def string_to_char_array(seq):
        """
        Converts an ASCII string to a NumPy array of byte-long ASCII codes.
        e.g. "ACGT" becomes [65, 67, 71, 84].
        """
        return np.frombuffer(bytearray(seq, "utf8"), dtype=np.int8)


    def char_array_to_string(arr):
        """
        Converts a NumPy array of byte-long ASCII codes into an ASCII string.
        e.g. [65, 67, 71, 84] becomes "ACGT".
        """
        return arr.tostring().decode("ascii")


    def one_hot_to_tokens(one_hot):
        """
        Converts an L x D one-hot encoding into an L-vector of integers in the range
        [0, D], where the token D is used when the one-hot encoding is all 0. This
        assumes that the one-hot encoding is well-formed, with at most one 1 in each
        column (and 0s elsewhere).
        """
        tokens = np.tile(one_hot.shape[1], one_hot.shape[0])  # Vector of all D
        seq_inds, dim_inds = np.where(one_hot)
        tokens[seq_inds] = dim_inds
        return tokens


    def tokens_to_one_hot(tokens, one_hot_dim):
        """
        Converts an L-vector of integers in the range [0, D] to an L x D one-hot
        encoding. The value `D` must be provided as `one_hot_dim`. A token of D
        means the one-hot encoding is all 0s.
        """
        identity = np.identity(one_hot_dim + 1)[:, :-1]  # Last row is all 0s
        return identity[tokens]

    if type(seq) is str:
        arr = string_to_char_array(seq)
    elif type(seq) is np.ndarray and len(seq.shape) == 2:
        seq_len, one_hot_dim = seq.shape
        arr = one_hot_to_tokens(seq)
    else:
        raise ValueError("Expected string or one-hot encoded array")


    if not rng:
        rng = np.random.RandomState()
   
    # Get the set of all characters, and a mapping of which positions have which
    # characters; use `tokens`, which are integer representations of the
    # original characters
    chars, tokens = np.unique(arr, return_inverse=True)

    # For each token, get a list of indices of all the tokens that come after it
    shuf_next_inds = []
    for t in range(len(chars)):
        mask = tokens[:-1] == t  # Excluding last char
        inds = np.where(mask)[0]
        shuf_next_inds.append(inds + 1)  # Add 1 for next token
 
    if type(seq) is str:
        all_results = []
    else:
        all_results = np.empty(
            (num_shufs if num_shufs else 1, seq_len, one_hot_dim),
            dtype=seq.dtype
        )

    for i in range(num_shufs if num_shufs else 1):
        # Shuffle the next indices
        for t in range(len(chars)):
            inds = np.arange(len(shuf_next_inds[t]))
            inds[:-1] = rng.permutation(len(inds) - 1)  # Keep last index same
            shuf_next_inds[t] = shuf_next_inds[t][inds]

        counters = [0] * len(chars)
       
        # Build the resulting array
        ind = 0
        result = np.empty_like(tokens)
        result[0] = tokens[ind]
        for j in range(1, len(tokens)):
            t = tokens[ind]
            ind = shuf_next_inds[t][counters[t]]
            counters[t] += 1
            result[j] = tokens[ind]

        if type(seq) is str:
            all_results.append(char_array_to_string(chars[result]))
        else:
            all_results[i] = tokens_to_one_hot(chars[result], one_hot_dim)
    return all_results if num_shufs else all_results[0]

